#include <iostream>
#include <stdlib.h>
#include "ros/ros.h"
#include <armadillo>
#include <udrone/forceCommand.h>
#include "qualisys/Subject.h"
#include "geometry_msgs/Pose.h"
#include "geometry_msgs/Wrench.h"
#include "udrone/xhat.h"

using namespace arma;

class EKF
{
public:
    EKF();
    void spin();
private:
    void controlForceUpdate(const udrone::forceCommand force);
    void QualisysPositionUpdate(const qualisys::Subject status);
    mat Smtrx(mat);

    ros::NodeHandle n;
    ros::Subscriber subControlForce;
    ros::Subscriber subQualisysPosition;
    ros::Publisher pubxhat;

    geometry_msgs::Pose QualisysPosition;
    udrone::forceCommand force;
    
    bool filter_on;

    // System properties
    float m;
    mat M;
    mat M_inv;
    mat D_NL;
    mat D_L;
    mat D;
    mat C_RB;
    mat C_A;
    mat C;
    float x_g, y_g, z_g;
    float X_du, Y_dv, Z_dw, N_dr;
    mat T_b_inv;
    mat u;

    // EKF properties
    mat R;
    mat Q;

    // System variables
    udrone::xhat xhat_out;
    mat xhat;
    mat xhat_apriori;
    mat y;
    mat eta;
    mat eta_apriori;
    mat nu;
    mat b;

    // EKF variables
    mat P;
    mat P_apriori;
    mat H;
    mat K;
    mat J_psi;
    mat f;
    mat df;
    mat A;
    mat B;
    mat E;
};

EKF::EKF()
{
    // System properties
    this->m              = 7.35;
    this->M              = zeros(4,4);
    this->D_NL           = zeros(4,4);
    this->D_L            = zeros(4,4);
    this->M.diag()       = vec({9.8528, 14.4901, 17.35, 0.204});
    this->M_inv          = inv(this->M);
    this->D_NL.diag()    = vec({58.34, 114.8, 93.42, 1.6256});
    this->D_L.diag()     = vec({0.708, 5.0, 4.847, 0.0666});
    this->C_RB           = zeros(4,4);
    this->C_A            = zeros(4,4);
    this->x_g            = 0;
    this->y_g            = 0;
    this->z_g            = 0.014;
    this->X_du           = -2.5028;
    this->Y_dv           = -7.1401;
    this->Z_dw           = -10;
    this->N_dr           = -0.0395;
    this->T_b_inv        = zeros(4,4);
    this->T_b_inv.diag() = vec({0.000001, 0.000001, 0.000001, 0.000001});
    this->u              = zeros(4,1);

    // Matrix setup
    this->eta            = zeros(5,1);
    this->eta_apriori    = zeros(5,1);
    this->P              = zeros(12,12);
    this->P_apriori      = eye(12,12);
    this->xhat           = zeros(12,1);
    this->xhat_apriori   = zeros(12,1);
    this->nu             = zeros(4,1);
    this->b              = zeros(4,1);
    this->y              = zeros(4,1);
    this->H              = eye(4,12);
    this->R              = zeros(4,4),
    this->Q              = zeros(12,12);
    this->J_psi          = eye(4,4);
    this->f              = zeros(12,1);
    this->df             = zeros(12,12);
    this->A              = eye(12,12);
    this->B              = zeros(12,4);
    this->B(span(4,7),span(0,3))   = this->M_inv;
    this->E              = zeros(12,12);
    this->E(span(4,7),span(4,7))   = 0.1 * this->M_inv;
    this->E(span(8,11),span(8,11)) = 0.1 * eye(4,4);


    // EKF tuning
    this->R.diag()       = vec({0.01, 0.01, 0.01, 0.01});
    this->Q.diag()       = vec({0.01, 0.01, 0.01, 0.01, 0.1, 0.1, 0.1, 0.1, 0.000001, 0.000001, 0.000001, 0.000001});



    this->subControlForce               = n.subscribe("forceCommand",10,&EKF::controlForceUpdate, this);
    this->subQualisysPosition           = n.subscribe("/qualisys/uDrone", 10, &EKF::QualisysPositionUpdate, this);
    this->pubxhat                       = n.advertise<udrone::xhat>("xhat",10);

    this->filter_on = true;
}


void EKF::controlForceUpdate(const udrone::forceCommand msg) {
    this->u(0,0) = msg.force.force.x;
    this->u(1,0) = msg.force.force.y;
    this->u(2,0) = msg.force.force.z;
    this->u(3,0) = msg.force.torque.z;
}

void EKF::QualisysPositionUpdate(const qualisys::Subject status) {
    mat orientation = zeros(4,1);

    orientation(0,0) = status.orientation.x;
    orientation(1,0) = status.orientation.y;
    orientation(2,0) = status.orientation.z;
    orientation(3,0) = status.orientation.w;

    double t3  = 2.0 * (orientation(3,0) * orientation(2,0) + orientation(0,0) * orientation(1,0));
	double t4  = 1.0 - 2.0 * (orientation(1,0) * orientation(1,0) + orientation(2,0) * orientation(2,0));  
	double psi = std::atan2(t3, t4);

    y(0,0) = status.position.x;
    y(1,0) = status.position.y;
    y(2,0) = status.position.z;
    y(3,0) = psi;

    if (this->filter_on) {
        // Update cycle
        this->K    = this->P_apriori * this->H.t() * inv(this->H * this->P_apriori * this->H.t() + R);
        this->xhat = this->xhat_apriori + K * (this->y - this->H * this->xhat_apriori);
        this->P    = (eye(12,12) - this->K * this->H) * this->P_apriori * trans(eye(12,12) - this->K * this->H) + this->K * this->R * this->K.t();

        // Predict cycle
        this->J_psi(0,0) = cos(this->xhat(3,0));
        this->J_psi(1,0) = sin(this->xhat(3,0));
        this->J_psi(0,1) =-sin(this->xhat(3,0));
        this->J_psi(1,1) = cos(this->xhat(3,0));

        this->D          = zeros(4,4);
        this->D(0,0)     = this->D_NL(0,0) * abs(this->xhat(4,0));
        this->D(1,1)     = this->D_NL(1,1) * abs(this->xhat(5,0));
        this->D(2,2)     = this->D_NL(2,2) * abs(this->xhat(6,0));
        this->D(3,3)     = this->D_NL(3,3) * abs(this->xhat(7,0));

        this->C_RB(0,3)  =-this->m * (this->x_g * this->xhat(3,0) + this->xhat(1,0));
        this->C_RB(1,3)  =-this->m * (this->y_g * this->xhat(3,0) - this->xhat(0,0));
        this->C_RB(3,0)  = this->m * (this->x_g * this->xhat(3,0) + this->xhat(1,0));
        this->C_RB(3,1)  = this->m * (this->y_g * this->xhat(3,0) - this->xhat(1,0));

        this->C_A(0,3)   = this->Y_dv * this->xhat(5,0);
        this->C_A(1,3)   =-this->X_du * this->xhat(4,0);
        this->C_A(3,0)   =-this->Y_dv * this->xhat(5,0);
        this->C_A(3,1)   = this->X_du * this->xhat(4,0);

        this->C          = this->C_A + this->C_RB;

        this->f(span(0,3), 0) = this->J_psi * this->xhat(span(4,7), 0);
        this->f(span(4,7), 0) =-this->M_inv * this->D * this->xhat(span(4,7), 0) - this->M_inv * this->D_L * this->xhat(span(4,7), 0) - this->M_inv * this->C * this->xhat(span(4,7), 0) + this->M_inv * this->J_psi.t() * this->xhat(span(8,11), 0);
        this->f(span(8,11),0) =-this->T_b_inv * this->xhat(span(8,11),0);

        this->df(span(0,3), span(4,7))  = this->J_psi;
        this->df(span(4,7), span(4,7))  =-this->M_inv * this->D - this->M_inv * this->D_L - this->M_inv * this->C;
        this->df(span(4,7), span(8,11)) = this->M_inv * this->J_psi.t();
        this->df(span(8,11),span(8,11)) =-this->T_b_inv;

        this->A = eye(12,12) + this->df * 0.1;

        this->xhat_apriori = this->xhat + 0.1 * this->f + this->B * this->u;
        this->P_apriori    = this->A * this->P * this->A.t() + this->E * this->Q * this->E.t();

        this->xhat_out.x   = this->xhat(0,0);
        this->xhat_out.y   = this->xhat(1,0);
        this->xhat_out.z   = this->xhat(2,0);
        this->xhat_out.psi = this->xhat(3,0);
    } else {
        this->xhat_out.x   = y(0,0);
        this->xhat_out.y   = y(1,0);
        this->xhat_out.z   = y(2,0);
        this->xhat_out.psi = y(3,0);
    }

    this->pubxhat.publish(this->xhat_out);
}

mat EKF::Smtrx(mat r) {
    mat S = zeros(3,3);
    S.row(0) = rowvec({      0, -r(2,0),  r(1,0)});
    S.row(1) = rowvec({ r(2,0),       0, -r(0,0)});
    S.row(2) = rowvec({-r(1,0),  r(0,0),       0});

    return S;
}

void EKF::spin()
{
    float loop_periode = 0.1;

    ros::Rate loop(1/loop_periode);
    while(ros::ok()) {
        ros::spinOnce();
        loop.sleep();
    }
}

int main(int argc, char **argv)
{
    ros::init(argc, argv, "EKF");
    EKF EKF;
    EKF.spin();
    return 0;
}

